/* symbols.c - contains functions related to the processing of symbol
 * table entries.
 * Written by Owen Klan  -  October 2002 */

/* Update History:
 * 4th October 2002  -  Began modifications for 0.2 tree. Mostly involving
 *                      byte ordering for 16 and 32 bit data items
 * 25th October 2002 -  Added symbol table reporting in columned format
 *                      as well as listing for function information
 */

#include <stdio.h>
#include <elf.h>
#include <stdlib.h>
#include <string.h>

#include "elfdump.h"

/* function prototypes */
static void dump_symbol(Elf32_Sym *symbol, int n);
static void dump_export(Elf32_Sym *symbol);
static char *get_symbol_type(Elf32_Sym *symbol);
char *get_sym_type(Elf32_Sym *symbol);
char *get_sym_bind(Elf32_Sym *symbol);

/* reads and displays information from the symbol table */
void display_symbols(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    int i = 0;
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;		       /* current symbol */
    Elf32_Sym *symbol;

    /* if no string table can be found for the symbols then die */
    if (sym_name_table == NULL) {
	fprintf(stdout, "Symbol table not found, possibly a stripped"
		" binary.\n");
	return;
    }
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum) * ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (ptr->sh_type != SHT_SYMTAB) {
	ptr++;
    }

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    (void *)symbol = (void *)sym_tab;
    
    /* now process all of the symbols */
    fprintf(stdout, "\nSymbol Information\n");
    
    for (i=0; i<num_syms; i++) {
	dump_symbol(symbol, i);
	symbol++;
    }
}


/* reads and displays information from the symbol table in 
 * columned format */
void symbols_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    int i = 0;
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;	       /* current symbol */
    Elf32_Sym *symbol;
    char *type_str, *bind_str;
    char slink_str[8];		       /* String for symbol link */

    /* if no string table can be found for the symbols then die */
    if (sym_name_table == NULL) {
	fprintf(stdout, "Symbol table not found, possibly a stripped"
		" binary.\n");
	return;
    }
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum) * ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (ptr->sh_type != SHT_SYMTAB) {
	ptr++;
    }

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    (void *)symbol = (void *)sym_tab;
    
    /* now process all of the symbols */
    fprintf(stdout, "\nSymbol Information\n");
    fprintf(stdout, "Symbol  Value     %6s  Type  Binding  %8s  "
	    "Name\n", "Size", "Sec.Link");
    for (i=0; i<num_syms; i++) {
	type_str = get_sym_type(symbol);
	bind_str = get_sym_bind(symbol);
	
	if (ELF32_ST_TYPE(symbol->st_info) == STT_OBJECT
	 || ELF32_ST_TYPE(symbol->st_info) == STT_FILE) {
	    sprintf(slink_str, "NA");
	} else {
	    sprintf(slink_str, "%d", ehalf_swap(symbol->st_shndx));
	}
	
	fprintf(stdout, " %4d   %08X  %6X  %s   %3s       %2s      %s\n", i,
		eword_swap(symbol->st_value), eword_swap(symbol->st_size),
		type_str, bind_str, slink_str, 
		(sym_name_table + eword_swap(symbol->st_name)));
	symbol++;
	if (type_str)
	  free(type_str);
	if (bind_str)
	  free(bind_str);
    }
}

/* Procedure to display exported function and variable names */
void display_exports(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    int i = 0;
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;		       /* current symbol */
    Elf32_Sym *symbol;

    /* if no string table can be found for the symbols then die */
    if (sym_name_table == NULL) {
	fprintf(stdout, "Symbol table not found, possibly a stripped"
		" binary.\n");
	return;
    }
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum) * ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (eword_swap(ptr->sh_type) != SHT_SYMTAB) {
	ptr++;
    }

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    (void *)symbol = (void *)sym_tab;
    
    /* now process all of the symbols, showing global function
     and data items */
    fprintf(stdout, "\nExports Information\n");
    
    for (i=0; i<num_syms; i++) {
	if (ELF32_ST_BIND(symbol->st_info) == STB_GLOBAL)
	  dump_export(symbol);
	symbol++;
    }
}

/* Procedure to display information regarding exported symbols */
static void dump_export(Elf32_Sym *symbol)
{
    if (ELF32_ST_TYPE(symbol->st_info) == STT_FUNC) {
	fprintf(stdout, "\n Function:     %s\n",
		(sym_name_table + eword_swap(symbol->st_name)));
	fprintf(stdout, " Address:      0x%08X\n",
		eword_swap(symbol->st_value));
    } else if (ELF32_ST_TYPE(symbol->st_info) == STT_OBJECT) {
	fprintf(stdout, "\n Data Object:  %s\n",
		(sym_name_table + eword_swap(symbol->st_name)));
	fprintf(stdout, " Size:  %d\t\tValue:  0x%08X (%d)\n",
		eword_swap(symbol->st_size), eword_swap(symbol->st_value),
		eword_swap(symbol->st_value));
    }
}

/* Procedure that extracts function symbols from the symbol table
 * and displays the output in a columned format */
void functions_report(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    int i = 0;
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;	       /* current symbol */
    Elf32_Sym *symbol;
    
    /* if no string table can be found for the symbols then die */
    if (sym_name_table == NULL) {
	fprintf(stdout, "Symbol table not found, possibly a stripped"
		" binary.\n");
	return;
    }
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum) * ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (eword_swap(ptr->sh_type) != SHT_SYMTAB) {
	ptr++;
    }

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    (void *)symbol = (void *)sym_tab;
    
    /* now process all of the symbols, showing function items only */
    fprintf(stdout, "\nFunctions Information\n");
    fprintf(stdout, "%-40s   %-8s %-8s %-8s\n",
	    "Function", "Address", "Binding", "Sec.Link");
    for (i=0; i < num_syms; i++) {
	char *bind_str = NULL;

	if (ELF32_ST_TYPE(symbol->st_info) == STT_FUNC) {
	    bind_str = get_sym_bind(symbol);
	    
	    /* Display the function information */
	    fprintf(stdout, "%-40s   %08X   %-4s    %d\n",
		    (sym_name_table + eword_swap(symbol->st_name)),
		    eword_swap(symbol->st_value), bind_str,
		    ehalf_swap(symbol->st_shndx));
	    
	    if (bind_str)
	      free(bind_str);
	}
	symbol++;
    }
}


/* Procedure to extract filename symbols from the symbol table */
void display_files(Elf32_Ehdr *header, char *image)
{
    Elf32_Shdr *sectab;                /* section table offset */
    Elf32_Half sectab_size;            /* section table size */
    int num_syms;		       /* number of symbols */
    int i = 0;
    Elf32_Shdr *ptr;                   /* segment table buffer */
    Elf32_Shdr *sym_tab;		       /* current symbol */
    Elf32_Sym *symbol;

    /* if no string table can be found for the symbols then die */
    if (sym_name_table == NULL) {
	fprintf(stdout, "Symbol table not found, possibly a stripped"
		" binary.\n");
	return;
    }
    
    /* find the symbol table section */
    sectab = (Elf32_Shdr *)(image + eword_swap(header->e_shoff));
    sectab_size = ehalf_swap(header->e_shnum)* ehalf_swap(header->e_shentsize);
        
    ptr = (Elf32_Shdr *)sectab;
    while (eword_swap(ptr->sh_type) != SHT_SYMTAB) {
	ptr++;
    }

    num_syms = eword_swap(ptr->sh_size) / eword_swap(ptr->sh_entsize);
    sym_tab = (Elf32_Shdr *)(image + eword_swap(ptr->sh_offset));
    (void *)symbol = (void *)sym_tab;
    
    /* now process all of the symbols, displaying only those of
     * type STT_FILE */
    fprintf(stdout, "\nFiles Information\n\n");

    for (i=0; i<num_syms; i++) {
	if (ELF32_ST_TYPE(symbol->st_info) == STT_FILE) {
	    fprintf(stdout, " %s\n",
		    (sym_name_table + eword_swap(symbol->st_name)));
	}
	symbol++;
    }
}

/* Procedure to display information about a program symbol */
static void dump_symbol(Elf32_Sym *symbol, int n)
{
    char *type_str = NULL;
    
    if (symbol == NULL)
      return;
    
    type_str = get_symbol_type(symbol);
    
    if (symbol->st_name != 0 && sym_name_table != NULL) {
	fprintf(stdout, "\n %s\n", 
		(sym_name_table + eword_swap(symbol->st_name)));
	fprintf(stdout, "  Type:  %s\t\tValue:  0x%08X\n",
		type_str, eword_swap(symbol->st_value));
    }
    
    if (type_str)
      free(type_str);
}

/* Function that returns a string describing a symbol's type. */
static char *get_symbol_type(Elf32_Sym *symbol)
{
    char *sym_type;
    
    /* allocate and clear some string space */
    if ((sym_type = malloc(32)) == NULL)
      return NULL;
    memset(sym_type, 0x00, 32);
    
    switch (ELF32_ST_TYPE(symbol->st_info)) {
      case STT_NOTYPE:
	strncpy(sym_type, "No type", 31); break;
      case STT_OBJECT:
	strncpy(sym_type, "Data object", 31); break;
      case STT_FUNC:
	strncpy(sym_type, "Function", 31); break;
      case STT_SECTION:
	strncpy(sym_type, "Section", 31); break;
      case STT_FILE:
	strncpy(sym_type, "File", 31); break;
      default:
	sprintf(sym_type, "Unknown"); break;
    }
    
    return sym_type;
}

/* Function that returns a 4 character code specifying a symbols type.
 * Note that this, unlike most functions at this stage, DOESN'T use
 * static data! Whoo Hoo! */
char *get_sym_type(Elf32_Sym *symbol)
{
    char *typestr;
    
    if ((typestr = malloc(5)) == NULL)
      return NULL;
    else
      memset(typestr, 0x00, 5);
    
    switch (ELF32_ST_TYPE(symbol->st_info)) {
      case STT_NOTYPE: sprintf(typestr, "NONE"); break;
      case STT_OBJECT: sprintf(typestr, "OBJ "); break;
      case STT_FUNC:   sprintf(typestr, "FUNC"); break;
      case STT_SECTION: sprintf(typestr, "SECT"); break;
      case STT_FILE:   sprintf(typestr, "FILE"); break;
      default:         sprintf(typestr, "????");
    }
    
    return typestr;
}

/* Function that returns a 4 character code specifying a symbols bindings.
 * Note that this, unlike most functions at this stage, DOESN'T use
 * static data! Whoo Hoo! */
char *get_sym_bind(Elf32_Sym *symbol)
{
    char *bindstr;
    
    if ((bindstr = malloc(5)) == NULL)
      return NULL;
    else
      memset(bindstr, 0x00, 5);
    
    switch (ELF32_ST_BIND(symbol->st_info)) {
      case STB_LOCAL:  sprintf(bindstr, "LOCL"); break;
      case STB_GLOBAL: sprintf(bindstr, "GLOB"); break;
      case STB_WEAK:   sprintf(bindstr, "WEAK"); break;
      default:         sprintf(bindstr, "????");
    }
    
    return bindstr;
}
